import os
import shutil
import json  
import pandas
import spacy
from spacy import displacy
from spacy.matcher import Matcher

def load_data(file_name):
    """Reads text data from a file, each line as a separate entry.
    
       Parameters
       ----------
        filename : str
            The file location of the data.

        Returns
        -------
        List[str]: 
            A list of strings, each string is a non-empty, whitespace-stripped line of text from the file.
    """

    with open(file_name, 'r', encoding='utf-8') as file:
        return [line.strip() for line in file if line.strip()]

def write_results(results_path, results):
    """Writes the results to JSON and CSV files.

        Parameters
        ----------
        results_path : str 
            The directory path where the results files will be saved.
        results : list of dict 
            A list of dictionaries containing the results data.

        Returns
        -------
        None
    """
    
    os.makedirs(results_path, exist_ok=True)
    json_path = os.path.join(results_path, 'results.json')
    csv_path = os.path.join(results_path, 'results.csv')

    with open(json_path, 'w', encoding='utf-8') as json_file:
        json.dump(results, json_file, indent=4)

    results_data_frame = pandas.read_json(json_path)
    results_data_frame.to_csv(csv_path, encoding = 'utf-8', index = False)
 
    print(f"Results saved: to {json_path} and {csv_path}") 

def extract_product_id(matcher, doc):  
    """Extracts the product_id from the text based on NER combined with POS and dependency patterns assuming there is only one product per entry.
        
        Parameters
        ----------
        matcher : spacy.matcher.Matcher
            The spaCy matcher object for matching patterns in the text.
        doc : spacy.tokens.Doc
            The spaCy document object containing the parsed text.

        Returns
        -------
        tuple : product_id (str) , extraction_method (str)
            A tuple containing the extracted product_id (str) and the extraction method (str).
    """
    product_id = None
    product_id_entity = None
    extraction_method = ""
  
    #get PRODUCT from NER
    for entity in doc.ents:
        if entity.label_ in ['PRODUCT']:
            product_id_entity = entity
            extraction_method = 'NER'
            break

    #clean up the named entities using POS
    if product_id_entity:
        buffer = []
        for product_id_token in doc[product_id_entity.start:product_id_entity.end]:
            if product_id_token.pos_ in ['DET']:
                continue
            if product_id_token.pos_ in ['ADP', 'PUNCT']: 
                break
            buffer.append(product_id_token.text)       
        product_id = ' '.join(buffer) 

    #if no product id was discovered, try to extract the id via pos/dependency matching  
    if not product_id_entity:
        product_id = extract_product_id_using_matcher(matcher, doc)
        extraction_method = 'MATCHER'
    if not product_id:
        product_id = extract_product_id_using_pos_and_deps(doc) 
        extraction_method = 'POS_AND_DEPS'     
 
    return product_id, extraction_method  

def extract_product_id_using_matcher(matcher, doc):
    """Extracts the product_id from the text using the provided matcher object. 
    
        Parameters
        ----------
        matcher : spacy.matcher.Matcher
            The spaCy matcher object for matching patterns in the text.
        doc : spacy.tokens.Doc
            The spaCy document object containing the parsed text.

        Returns
        -------
        str
            The extracted product_id or None if no match is found.
    """
    
    # Define a patterns for typical product names with trailing model numbers
    propn_sequence_1 = [
      {"POS": "NOUN", "OP": "?"}, 
      {"POS": "PROPN", "OP": "{2,}"},
      {"POS": "NUM", "OP": "?"}
    ]

    propn_sequence_2 = [
      {"POS": "NOUN", "OP": "+"}, 
      {"POS": "PROPN", "OP": "{2,}"},
      {"POS": "NUM", "OP": "?"}
    ]

    propn_sequence_3 = [
      {'POS': 'NOUN'},
      {'POS': 'PROPN', 'OP': '+'},
      {'POS': 'NOUN', 'OP': '?'}
    ]

    matcher.add("PRODUCT_PATTERN", [propn_sequence_1, propn_sequence_2, propn_sequence_3])

    matches = matcher(doc)
    product_id = None

    for match_id, start, end in matches:
        span = doc[start:end]  # The matched span
        product_id = span.text
    
    return product_id

def extract_product_id_using_pos_and_deps(doc):
    """Extracts the product_id from the text based on POS and dependency patterns.
    
        Parameters
        ----------
        doc : spacy.tokens.Doc
            The spaCy document object containing the parsed text.

        Returns
        -------
        str
            The extracted product_id or None if no match is found.
    """

    product_id = None
    buffer = []
    previous_token_dep = None  # Track the dependency of the previous token

    for token in doc:
        if token.pos_ in ['NOUN', 'PROPN']:    
            if token.dep_ in ['compound', 'amod', 'nsubj', 'pobj']:
                buffer.append(token.text)
                previous_token_dep = token.dep_  
            elif buffer:
                if token.pos_ in ['NUM', 'PROPN']:
                    buffer.append(token.text)
                    product_id = ' '.join(buffer)
                    if token.dep_ in ['dobj'] and previous_token_dep in ['compound']: 
                        continue
                    buffer = []
                else:
                    # For trailing model numbers, check if the current token is a NUM or PROPN with nummod and follows a dobj
                    if token.dep_ == 'nummod' and previous_token_dep == 'dobj':
                        buffer.append(token.text) 
                    else:
                        buffer = [] 
        
    if buffer and not product_id:
        product_id = ' '.join(buffer)

    return product_id

def save_dependency_tree_as_HTML(file_path, file_name, doc):
     """Saves the dependency tree visualization of the given document as an HTML file.
        
        Parameters
        ----------
        file_path : str
            The directory path where the HTML file will be saved.
        file_name : str
            The name of the HTML file (without extension).
        doc : spacy.tokens.Doc
            The spaCy document object containing the parsed text.

        Returns
        -------
        None
    """
    options = {"compact": True, "bg": "#ffffff", "color": "black", "font": "Source Sans Pro"}
    html = displacy.render(doc, style='dep', options = options)
    html_file_path = os.path.join(file_path, f"{file_name}.html")
    with open(html_file_path, 'w', encoding='utf-8') as html_file:
        html_file.write(html)    

def save_entity_recognition_as_HTML(file_path, file_name, doc):
     """Saves the named entity recognition (NER) visualization of the given document as an HTML file.
    
        Parameters
        ----------
        file_path : str
            The directory path where the HTML file will be saved.
        file_name : str
            The name of the HTML file (without extension).
        doc : spacy.tokens.Doc
            The spaCy document object containing the parsed text.

        Returns
        -------
        None
    """
    html = displacy.render(doc, style='ent', page=True)
    html_file_path = os.path.join(file_path, f"{file_name}.html")
    with open(html_file_path, 'w', encoding='utf-8') as html_file:
        html_file.write(html)                    

def clear_directory(directory):
    """Clears the contents of the specified directory and recreates it.
        
        Parameters
        ----------
        directory : str
            The directory path to be cleared.

        Returns
        -------
        None
    """
    if os.path.exists(directory):
        shutil.rmtree(directory)  
    os.makedirs(directory, exist_ok=True)  

def analyze_results(nlp, results, expected_ids, save_visualizations):
    """Analyzes the extracted results, compares them with expected IDs, and optionally saves visualizations.
    
        Parameters
        ----------
        nlp : spacy.language.Language
            The spaCy language model object.
        results : list
            A list of dictionaries containing the extracted results.
        expected_ids : list
            A list of expected product IDs for comparison.
        save_visualizations : bool
            A flag indicating whether to save visualizations of the results.

        Returns
        -------
        None
    """
    total_matches = 0
    total_entries = len(results)
    extraction_method_counts = {}
    
    for index, entry in enumerate(results, start=0):
        entry['expectedId'] = expected_ids[index]
        entry['matches'] = entry['productId'] == entry['expectedId'] 

        if (entry['matches']):
            total_matches += 1
        else:
            print(f"No match for Entry {entry['entryId']}: Found Product ID: {entry['productId']} , Expected ID: {entry['expectedId']} , Extraction method: {entry['extractionMethod']}")  

        extraction_method_counts[entry['extractionMethod']] = extraction_method_counts.get(entry['extractionMethod'], 0) + 1

    if(save_visualizations):
        matches_dir = os.path.join(os.path.dirname(__file__), "visualisations", "matches")
        mismatches_dir = os.path.join(os.path.dirname(__file__), "visualisations", "mismatches")
        clear_directory(mismatches_dir)
        clear_directory(matches_dir)
        
        for index, entry in enumerate(results): 
            doc = nlp(entry['originalText'])
            visualization_path = matches_dir if entry['matches'] else mismatches_dir
            save_dependency_tree_as_HTML(visualization_path, f"{index}_dep_tree", doc)
            save_entity_recognition_as_HTML(visualization_path, f"{index}_NER", doc)

    print(f"Extraction Methods in use: {extraction_method_counts}")

    accuracy_percentage = (total_matches / total_entries) * 100
    print(f"Accuracy: {accuracy_percentage:.2f}% ({total_matches}/{total_entries} correct matches)")

def extract_information(nlp, data):
    """Extracts specific product information from a text entry.
    
        Parameters
        ----------
        nlp : spacy.language.Language
            The spaCy language model object.
        data : list
            A list of text entries from which to extract information.

        Returns
        -------
        list : dict
            A list of dictionaries containing the extracted information for each text entry.
    """
    matcher = Matcher(nlp.vocab)
    results = []
    for index, text in enumerate(data, start=1):
        doc = nlp(text)
        product_id, extraction_method = extract_product_id(matcher, doc)
        info = {
            "entryId": index,
            "originalText": doc.text,
            "extractionMethod" : extraction_method,
            "productId": product_id
        }
        results.append(info)
    return results    

def main():
    nlp = spacy.load('en_core_web_trf')

    data = load_data(os.path.join(os.path.dirname(__file__), 'SensorDataComplex.txt'))
    expected_ids = load_data(os.path.join(os.path.dirname(__file__), 'ProductIDs.txt'))

    results = extract_information(nlp, data)
    analyze_results(nlp, results, expected_ids, True)
   
    write_results(os.path.dirname(__file__), results)

if __name__ == "__main__":
    main()
